<?php

declare(strict_types = 1);

namespace Drupal\tailwind_jit;

use Drupal\Core\Cache\CacheBackendInterface;
use Drupal\Core\Site\Settings;
use Drupal\Component\Datetime\TimeInterface;

/**
 * Cache service for the Tailwind JIT compiler.
 *
 * The default size of this cache bin (no. of rows in the database) is 5000.
 * It can be overruled in settings.php like other cache bins, e.g. to set an unlimited size:
 * $settings['database_cache_max_rows']['bins']['tailwind_jit'] = -1;
 */
class Cache {
   /**
   * The name of the settings variable controlling the cache lifetime.
   *
   * It can be overruled by adding this key to settings.php, e.g. to limit cache lifetime to one hour:
   * $settings['tailwind_jit_cache_lifetime'] = 3600;
   *
   * @var string
   */
  protected const SETTINGS_KEY_CACHE_LIFETIME = 'tailwind_jit_cache_lifetime';

  /**
   * The constructor.
   *
   * @param \Drupal\Core\Cache\CacheBackendInterface $cache
   *   The database cache bin for tailwind_jit module.
   * @param \Drupal\Component\Datetime\TimeInterface $time
   *   The time service.
   */
  public function __construct(
    protected CacheBackendInterface $cache,
    protected TimeInterface $time,
  ) { }

  /**
   * Returns the lifetime of the Tailwind JIT cache in seconds.
   *
   * A value of -1 means unlimited lifetime.
   *
   * @return int
   */
  public static function getLifetime(): int {
    return Settings::get(self::SETTINGS_KEY_CACHE_LIFETIME, CacheBackendInterface::CACHE_PERMANENT);
  }

  /**
   * Cache setter with implicit calculation of the expire timestamp.
   */
  public function set(string $cacheKey, string $compiledCss): void {
    [$themeName, $responseType] = explode(':', $cacheKey);
    $lifetime = (int) $this->getLifetime();
    $expires = ($lifetime == CacheBackendInterface::CACHE_PERMANENT) ? CacheBackendInterface::CACHE_PERMANENT : $this->time->getRequestTime() + $lifetime;
    $this->cache->set(
      $cacheKey,
      $compiledCss,
      $expires,
    );
  }

  public function get(string $cacheKey): bool|object {
    return $this->cache->get($cacheKey);
  }
}
